@@ -1,6 +1,6 @@
 // +build windows
 
-// Copyright 2019 The TCell Authors
+// Copyright 2020 The TCell Authors
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use file except in compliance with the License.
@@ -18,6 +18,9 @@ package tcell
 
 import (
 	"errors"
+	"fmt"
+	"os"
+	"strings"
 	"sync"
 	"syscall"
 	"unicode/utf16"
@@ -36,6 +39,8 @@ type cScreen struct {
 	style      Style
 	clear      bool
 	fini       bool
+	vten       bool
+	truecolor  bool
 
 	w int
 	h int
@@ -45,7 +50,8 @@ type cScreen struct {
 	oimode  uint32
 	oomode  uint32
 	cells   CellBuffer
-	colors  map[Color]Color
+
+	finiOnce sync.Once
 
 	sync.Mutex
 }
@@ -90,10 +96,13 @@ var winColors = map[Color]Color{
 	ColorWhite:   ColorWhite,
 }
 
-var k32 = syscall.NewLazyDLL("kernel32.dll")
+var (
+	k32 = syscall.NewLazyDLL("kernel32.dll")
+	u32 = syscall.NewLazyDLL("user32.dll")
+)
 
-// We have to bring in the kernel32.dll directly, so we can get access to some
-// system calls that the core Go API lacks.
+// We have to bring in the kernel32 and user32 DLLs directly, so we can get
+// access to some system calls that the core Go API lacks.
 //
 // Note that Windows appends some functions with W to indicate that wide
 // characters (Unicode) are in use.  The documentation refers to them
@@ -114,6 +123,7 @@ var (
 	procSetConsoleWindowInfo       = k32.NewProc("SetConsoleWindowInfo")
 	procSetConsoleScreenBufferSize = k32.NewProc("SetConsoleScreenBufferSize")
 	procSetConsoleTextAttribute    = k32.NewProc("SetConsoleTextAttribute")
+	procMessageBeep                = u32.NewProc("MessageBeep")
 )
 
 const (
@@ -121,6 +131,22 @@ const (
 	w32WaitObject0 = uintptr(0)
 )
 
+const (
+	// VT100/XTerm escapes understood by the console
+	vtShowCursor = "\x1b[?25h"
+	vtHideCursor = "\x1b[?25l"
+	vtCursorPos  = "\x1b[%d;%dH" // Note that it is Y then X
+	vtSgr0       = "\x1b[0m"
+	vtBold       = "\x1b[1m"
+	vtUnderline  = "\x1b[4m"
+	vtBlink      = "\x1b[5m" // Not sure this is processed
+	vtReverse    = "\x1b[7m"
+	vtSetFg      = "\x1b[38;5;%dm"
+	vtSetBg      = "\x1b[48;5;%dm"
+	vtSetFgRGB   = "\x1b[38;2;%d;%d;%dm" // RGB
+	vtSetBgRGB   = "\x1b[48;2;%d;%d;%dm" // RGB
+)
+
 // NewConsoleScreen returns a Screen for the Windows console associated
 // with the current process.  The Screen makes use of the Windows Console
 // API to display content and read events.
@@ -145,6 +171,24 @@ func (s *cScreen) Init() error {
 	}
 	s.out = out
 
+	s.truecolor = true
+
+	// ConEmu handling of colors and scrolling when in terminal
+	// mode is extremely problematic at the best.  The color
+	// palette will scroll even though characters do not, when
+	// emiting stuff for the last character.  In the future we
+	// might change this to look at specific versions of ConEmu
+	// if they fix the bug.
+	if os.Getenv("ConEmuPID") != "" {
+		s.truecolor = false
+	}
+	switch os.Getenv("TCELL_TRUECOLOR") {
+	case "disable":
+		s.truecolor = false
+	case "enable":
+		s.truecolor = true
+	}
+
 	cf, _, e := procCreateEvent.Call(
 		uintptr(0),
 		uintptr(1),
@@ -167,8 +211,23 @@ func (s *cScreen) Init() error {
 	s.resize()
 
 	s.fini = false
-	s.setInMode(modeResizeEn)
-	s.setOutMode(0)
+	s.setInMode(modeResizeEn | modeExtndFlg)
+
+	// 24-bit color is opt-in for now, because we can't figure out
+	// to make it work consistently.
+	if s.truecolor {
+		s.setOutMode(modeVtOutput | modeNoAutoNL | modeCookedOut)
+		var omode uint32
+		s.getOutMode(&omode)
+		if omode&modeVtOutput == modeVtOutput {
+			s.vten = true
+		} else {
+			s.truecolor = false
+		}
+	} else {
+		s.setOutMode(0)
+	}
+
 	s.clearScreen(s.style)
 	s.hideCursor()
 	s.Unlock()
@@ -187,15 +246,20 @@ func (s *cScreen) EnableMouse() {
 }
 
 func (s *cScreen) DisableMouse() {
-	s.setInMode(modeResizeEn)
+	s.setInMode(modeResizeEn | modeExtndFlg)
 }
 
 func (s *cScreen) Fini() {
+	s.finiOnce.Do(s.finish)
+}
+
+func (s *cScreen) finish() {
 	s.Lock()
 	s.style = StyleDefault
 	s.curx = -1
 	s.cury = -1
 	s.fini = true
+	s.vten = false
 	s.Unlock()
 
 	s.setCursorInfo(&s.ocursor)
@@ -261,12 +325,25 @@ type rect struct {
 	bottom int16
 }
 
+func (s *cScreen) emitVtString(vs string) {
+	esc := utf16.Encode([]rune(vs))
+	syscall.WriteConsole(s.out, &esc[0], uint32(len(esc)), nil, nil)
+}
+
 func (s *cScreen) showCursor() {
-	s.setCursorInfo(&cursorInfo{size: 100, visible: 1})
+	if s.vten {
+		s.emitVtString(vtShowCursor)
+	} else {
+		s.setCursorInfo(&cursorInfo{size: 100, visible: 1})
+	}
 }
 
 func (s *cScreen) hideCursor() {
-	s.setCursorInfo(&cursorInfo{size: 1, visible: 0})
+	if s.vten {
+		s.emitVtString(vtHideCursor)
+	} else {
+		s.setCursorInfo(&cursorInfo{size: 1, visible: 0})
+	}
 }
 
 func (s *cScreen) ShowCursor(x, y int) {
@@ -294,11 +371,6 @@ func (s *cScreen) HideCursor() {
 	s.ShowCursor(-1, -1)
 }
 
-type charInfo struct {
-	ch   uint16
-	attr uint16
-}
-
 type inputRecord struct {
 	typ  uint16
 	_    uint16
@@ -566,10 +638,10 @@ func (s *cScreen) getConsoleInput() error {
 				for krec.repeat > 0 {
 					// convert shift+tab to backtab
 					if mod2mask(krec.mod) == ModShift && krec.ch == vkTab {
-						s.PostEvent(NewEventKey(KeyBacktab, 0,
+						s.PostEventWait(NewEventKey(KeyBacktab, 0,
 							ModNone, ""))
 					} else {
-						s.PostEvent(NewEventKey(KeyRune, rune(krec.ch),
+						s.PostEventWait(NewEventKey(KeyRune, rune(krec.ch),
 							mod2mask(krec.mod), ""))
 					}
 					krec.repeat--
@@ -582,7 +654,7 @@ func (s *cScreen) getConsoleInput() error {
 				return nil
 			}
 			for krec.repeat > 0 {
-				s.PostEvent(NewEventKey(key, rune(krec.ch),
+				s.PostEventWait(NewEventKey(key, rune(krec.ch),
 					mod2mask(krec.mod), ""))
 				krec.repeat--
 			}
@@ -596,14 +668,14 @@ func (s *cScreen) getConsoleInput() error {
 			mrec.flags = getu32(rec.data[12:])
 			btns := mrec2btns(mrec.btns, mrec.flags)
 			// we ignore double click, events are delivered normally
-			s.PostEvent(NewEventMouse(int(mrec.x), int(mrec.y), btns,
+			s.PostEventWait(NewEventMouse(int(mrec.x), int(mrec.y), btns,
 				mod2mask(mrec.mod), ""))
 
 		case resizeEvent:
 			var rrec resizeRecord
 			rrec.x = geti16(rec.data[0:])
 			rrec.y = geti16(rec.data[2:])
-			s.PostEvent(NewEventResize(int(rrec.x), int(rrec.y)))
+			s.PostEventWait(NewEventResize(int(rrec.x), int(rrec.y)))
 
 		default:
 		}
@@ -625,6 +697,9 @@ func (s *cScreen) scanInput() {
 
 // Windows console can display 8 characters, in either low or high intensity
 func (s *cScreen) Colors() int {
+	if s.vten {
+		return 1 << 24
+	}
 	return 16
 }
 
@@ -670,10 +745,10 @@ func (s *cScreen) mapStyle(style Style) uint16 {
 	f, b, a := style.Decompose()
 	fa := s.oscreen.attrs & 0xf
 	ba := (s.oscreen.attrs) >> 4 & 0xf
-	if f != ColorDefault {
+	if f != ColorDefault && f != ColorReset {
 		fa = mapColor2RGB(f)
 	}
-	if b != ColorDefault {
+	if b != ColorDefault && b != ColorReset {
 		ba = mapColor2RGB(b)
 	}
 	var attr uint16
@@ -724,17 +799,55 @@ func (s *cScreen) GetContent(x, y int) (rune, []rune, Style, int) {
 	return mainc, combc, style, width
 }
 
+func (s *cScreen) sendVtStyle(style Style) {
+	esc := &strings.Builder{}
+
+	fg, bg, attrs := style.Decompose()
+
+	esc.WriteString(vtSgr0)
+
+	if attrs&(AttrBold|AttrDim) == AttrBold {
+		esc.WriteString(vtBold)
+	}
+	if attrs&AttrBlink != 0 {
+		esc.WriteString(vtBlink)
+	}
+	if attrs&AttrUnderline != 0 {
+		esc.WriteString(vtUnderline)
+	}
+	if attrs&AttrReverse != 0 {
+		esc.WriteString(vtReverse)
+	}
+	if fg.IsRGB() {
+		r, g, b := fg.RGB()
+		fmt.Fprintf(esc, vtSetFgRGB, r, g, b)
+	} else if fg.Valid() {
+		fmt.Fprintf(esc, vtSetFg, fg&0xff)
+	}
+	if bg.IsRGB() {
+		r, g, b := bg.RGB()
+		fmt.Fprintf(esc, vtSetBgRGB, r, g, b)
+	} else if bg.Valid() {
+		fmt.Fprintf(esc, vtSetBg, bg&0xff)
+	}
+	s.emitVtString(esc.String())
+}
+
 func (s *cScreen) writeString(x, y int, style Style, ch []uint16) {
 	// we assume the caller has hidden the cursor
 	if len(ch) == 0 {
 		return
 	}
-	nw := uint32(len(ch))
-	procSetConsoleTextAttribute.Call(
-		uintptr(s.out),
-		uintptr(s.mapStyle(style)))
 	s.setCursorPos(x, y)
-	syscall.WriteConsole(s.out, &ch[0], nw, &nw, nil)
+
+	if s.vten {
+		s.sendVtStyle(style)
+	} else {
+		procSetConsoleTextAttribute.Call(
+			uintptr(s.out),
+			uintptr(s.mapStyle(style)))
+	}
+	syscall.WriteConsole(s.out, &ch[0], uint32(len(ch)), nil, nil)
 }
 
 func (s *cScreen) draw() {
@@ -747,13 +860,13 @@ func (s *cScreen) draw() {
 	}
 	buf := make([]uint16, 0, s.w)
 	wcs := buf[:]
-	lstyle := Style(-1) // invalid attribute
+	lstyle := styleInvalid
 
 	lx, ly := -1, -1
 	ra := make([]rune, 1)
 
-	for y := 0; y < int(s.h); y++ {
-		for x := 0; x < int(s.w); x++ {
+	for y := 0; y < s.h; y++ {
+		for x := 0; x < s.w; x++ {
 			mainc, combc, style, width := s.cells.GetContent(x, y)
 			dirty := s.cells.Dirty(x, y)
 			if style == StyleDefault {
@@ -766,7 +879,7 @@ func (s *cScreen) draw() {
 				// cells, or because we need to change styles
 				s.writeString(lx, ly, lstyle, wcs)
 				wcs = buf[0:0]
-				lstyle = Style(-1)
+				lstyle = StyleDefault
 				if !dirty {
 					continue
 				}
@@ -793,7 +906,7 @@ func (s *cScreen) draw() {
 		}
 		s.writeString(lx, ly, lstyle, wcs)
 		wcs = buf[0:0]
-		lstyle = Style(-1)
+		lstyle = styleInvalid
 	}
 }
 
@@ -844,12 +957,18 @@ func (s *cScreen) setCursorInfo(info *cursorInfo) {
 	procSetConsoleCursorInfo.Call(
 		uintptr(s.out),
 		uintptr(unsafe.Pointer(info)))
+
 }
 
 func (s *cScreen) setCursorPos(x, y int) {
-	procSetConsoleCursorPosition.Call(
-		uintptr(s.out),
-		coord{int16(x), int16(y)}.uintptr())
+	if s.vten {
+		// Note that the string is Y first.  Origin is 1,1.
+		s.emitVtString(fmt.Sprintf(vtCursorPos, y+1, x+1))
+	} else {
+		procSetConsoleCursorPosition.Call(
+			uintptr(s.out),
+			coord{int16(x), int16(y)}.uintptr())
+	}
 }
 
 func (s *cScreen) setBufferSize(x, y int) {
@@ -888,7 +1007,6 @@ func (s *cScreen) resize() {
 		uintptr(s.out),
 		uintptr(1),
 		uintptr(unsafe.Pointer(&r)))
-
 	s.PostEvent(NewEventResize(w, h))
 }
 
@@ -905,32 +1023,50 @@ func (s *cScreen) Fill(r rune, style Style) {
 }
 
 func (s *cScreen) clearScreen(style Style) {
-	pos := coord{0, 0}
-	attr := s.mapStyle(style)
-	x, y := s.w, s.h
-	scratch := uint32(0)
-	count := uint32(x * y)
+	if s.vten {
+		s.sendVtStyle(style)
+		row := strings.Repeat(" ", s.w)
+		for y := 0; y < s.h; y++ {
+			s.setCursorPos(0, y)
+			s.emitVtString(row)
+		}
+		s.setCursorPos(0, 0)
 
-	procFillConsoleOutputAttribute.Call(
-		uintptr(s.out),
-		uintptr(attr),
-		uintptr(count),
-		pos.uintptr(),
-		uintptr(unsafe.Pointer(&scratch)))
-	procFillConsoleOutputCharacter.Call(
-		uintptr(s.out),
-		uintptr(' '),
-		uintptr(count),
-		pos.uintptr(),
-		uintptr(unsafe.Pointer(&scratch)))
+	} else {
+		pos := coord{0, 0}
+		attr := s.mapStyle(style)
+		x, y := s.w, s.h
+		scratch := uint32(0)
+		count := uint32(x * y)
+
+		procFillConsoleOutputAttribute.Call(
+			uintptr(s.out),
+			uintptr(attr),
+			uintptr(count),
+			pos.uintptr(),
+			uintptr(unsafe.Pointer(&scratch)))
+		procFillConsoleOutputCharacter.Call(
+			uintptr(s.out),
+			uintptr(' '),
+			uintptr(count),
+			pos.uintptr(),
+			uintptr(unsafe.Pointer(&scratch)))
+	}
 }
 
 const (
+	// Input modes
 	modeExtndFlg uint32 = 0x0080
-	modeMouseEn  uint32 = 0x0010
-	modeResizeEn uint32 = 0x0008
-	modeWrapEOL  uint32 = 0x0002
-	modeCooked   uint32 = 0x0001
+	modeMouseEn         = 0x0010
+	modeResizeEn        = 0x0008
+	modeCooked          = 0x0001
+	modeVtInput         = 0x0200
+
+	// Output modes
+	modeCookedOut uint32 = 0x0001
+	modeWrapEOL          = 0x0002
+	modeVtOutput         = 0x0004
+	modeNoAutoNL         = 0x0008
 )
 
 func (s *cScreen) setInMode(mode uint32) error {
@@ -1033,3 +1169,24 @@ func (s *cScreen) HasKey(k Key) bool {
 
 func (s *cScreen) RegisterRawSeq(string) {}
 func (s *cScreen) SetPaste(bool)         {}
+
+func (s *cScreen) GetClipboard(string) error {
+	return errors.New("Not supported on Windows")
+}
+
+func (s *cScreen) SetClipboard(string, string) error {
+	return errors.New("Not supported on Windows")
+}
+
+func (s *cScreen) Beep() error {
+	// A simple beep. If the sound card is not available, the sound is generated
+	// using the speaker.
+	//
+	// Reference:
+	// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-messagebeep
+	const simpleBeep = 0xffffffff
+	if rv, _, err := procMessageBeep.Call(simpleBeep); rv == 0 {
+		return err
+	}
+	return nil
+}
